#include <stdio.h>
#include <stdlib.h>
#include "calcul_de_triangles.h"

triangle* creer_triangle(sommet s0, sommet s1, sommet s2) {
    triangle* t = (triangle*) malloc(sizeof(triangle));
    t->s0 = s0;
    t->s1 = s1;
    t->s2 = s2;
    return t;
}

void detruire_triangle(triangle* t) {
    free(t);
}

liste_sommets creer_liste_sommets() {
    liste_sommets l = NULL;
    return l;
}

void detruire_liste_sommets(liste_sommets l) {
    if(l) {
        detruire_liste_sommets(l->suivant);
        free(l);
    }
}

uint longueur_liste_sommets(liste_sommets l) {
    uint longueur = 0;
    for (ml_sommets* m = l; m; m = m->suivant)
        ++longueur;
    return longueur;
}

void ajouter_sommet(liste_sommets* l, sommet s) {
    ml_sommets* m = (ml_sommets*) malloc(sizeof(ml_sommets));
    m->sommet = s;
    m->suivant = *l;
    *l = m;
}

void inverser_liste_sommets(liste_sommets* l) {
    if (!l || !(*l) || !(*l)->suivant)
        return;
    ml_sommets* p = *l;
    *l = (*l)->suivant;
    p->suivant = NULL;
    ml_sommets* tmp = (*l)->suivant;
    (*l)->suivant = p;
    while(tmp) {
        p = *l;
        *l = tmp;
        tmp = (*l)->suivant;
        (*l)->suivant = p;
    }
}

liste_triangles creer_liste_triangles() {
    liste_triangles l = NULL;
    return l;
}

void detruire_liste_triangles(liste_triangles l) {
    if(l) {
        detruire_liste_triangles(l->suivant);
        detruire_triangle(l->triangle);
        free(l);
    }
}

void ajouter_triangle(liste_triangles* l, triangle* t) {
    ml_triangles* m = (ml_triangles*) malloc(sizeof(ml_triangles));
    m->triangle = t;
    m->suivant = *l;
    *l = m;
}

liste_adjacence* creer_liste_adjacence(uint nb_sommets) {
    liste_adjacence* la = (liste_adjacence*) malloc(sizeof(liste_adjacence));
    la->nb_sommets = nb_sommets;
    la->voisins = (liste_sommets*) malloc(sizeof(liste_sommets) * nb_sommets);
    for (sommet i = 0; i < nb_sommets; ++i)
        la->voisins[i] = creer_liste_sommets();
    return la;
}

void detruire_liste_adjacence(liste_adjacence* l) {
    if(l) {
        for (sommet i = 0; i < l->nb_sommets; ++i)
            detruire_liste_sommets(l->voisins[i]);
        free(l);
    }
}

void ajouter_voisins_liste(liste_adjacence* l, sommet s0, sommet s1) {
    ajouter_sommet(&l->voisins[s0], s1);
    ajouter_sommet(&l->voisins[s1], s0);
}

matrice_adjacence* creer_matrice_adjacence(uint nb_sommets) {
    matrice_adjacence* ma = (matrice_adjacence*) malloc(sizeof(matrice_adjacence));
    ma->nb_sommets = nb_sommets;
    ma->matrice = (bool**) malloc(sizeof(bool*) * nb_sommets);
    for (sommet i = 0; i < nb_sommets; ++i) {
        ma->matrice[i] = (bool*) malloc(sizeof(bool) * nb_sommets);
        for (sommet j = 0; j < nb_sommets; ++j)
            ma->matrice[i][j] = false;
    }
    return ma;
}

void detruire_matrice_adjacence(matrice_adjacence* m) {
    if(m) {
        for (sommet i = 0; i < m->nb_sommets; ++i)
            free(m->matrice[i]);
        free(m);
    }
}

void ajouter_voisins_matrice(matrice_adjacence* m, sommet s0, sommet s1) {
    m->matrice[s0][s1] = true;
    m->matrice[s1][s0] = true;
}

liste_triangles calcul_triangle_1(matrice_adjacence* m) {
    liste_triangles lt = creer_liste_triangles();
    for (sommet i = 0 ; i < m->nb_sommets ; ++i)
        for (sommet j = i+1 ; j < m->nb_sommets ; ++j)
            for (sommet k = j+1 ; k < m->nb_sommets ; ++k)
                if ( m->matrice[i][j] && m->matrice[j][k] && m->matrice[k][i] )
                    ajouter_triangle(&lt, creer_triangle(i,j,k));
    return lt;
}

void intersection(liste_adjacence* la, liste_triangles* lt,
                  sommet i, sommet j) {
    liste_sommets li = la->voisins[i];
    liste_sommets lj = la->voisins[j];
    while( li && lj ) {
        if (li->sommet == lj->sommet) {
            if (li->sommet > j)
                ajouter_triangle(lt, creer_triangle(i,j,li->sommet));
            li = li->suivant;
            lj = lj->suivant;
        } else if (li->sommet > lj->sommet )
            lj = lj->suivant;
        else
            li = li->suivant;
    }
}

liste_triangles calcul_triangle_2(liste_adjacence* la) {
    liste_triangles lt = creer_liste_triangles();
    for (sommet i = 0 ; i < la->nb_sommets ; ++i)
        for (liste_sommets li = la->voisins[i] ; li ; li = li->suivant)
            if (li->sommet > i)
                intersection(la, &lt, i, li->sommet);
    return lt;
}

void lourd_leger(liste_adjacence* la, liste_sommets* ll, liste_sommets* lh) {
    *ll = creer_liste_sommets();
    *lh = creer_liste_sommets();
    uint longueur;
    for (sommet s = 0; s < la->nb_sommets ; ++s) {
        longueur = longueur_liste_sommets(la->voisins[s]);
        if (longueur * longueur < la->nb_sommets)
            ajouter_sommet(ll, s);
        else
            ajouter_sommet(lh, s);
    }
    inverser_liste_sommets(ll);
    inverser_liste_sommets(lh);
}

void liste_lourd_leger(liste_adjacence* la, liste_sommets ll, liste_sommets lh,
                       liste_adjacence* lal, liste_adjacence* lah) {
    for (sommet s = 0; s < la->nb_sommets ; ++s) {
        liste_sommets li = la->voisins[s], lhtmp = lh, lltmp = ll;
        while( li ) {
            bool suivant_h = true, suivant_l = true;
            if ( lhtmp ) {
                suivant_h = false;
                if (li->sommet == lhtmp->sommet) {
                    ajouter_sommet(&(lah->voisins[s]), lhtmp->sommet);
                    li = li->suivant;
                    lhtmp = lhtmp->suivant;
                    continue;
                } else if (li->sommet > lhtmp->sommet )
                    lhtmp = lhtmp->suivant;
                else
                    suivant_h = true;
            }
            if ( lltmp ) {
                suivant_l = false;
                if (li->sommet == lltmp->sommet) {
                    ajouter_sommet(&(lal->voisins[s]), lltmp->sommet);
                    li = li->suivant;
                    lltmp = lltmp->suivant;
                } else if (li->sommet > lltmp->sommet )
                    lltmp = lltmp->suivant;
                else
                    suivant_l = true;
            }
            if ( suivant_l && suivant_h )
                li = li->suivant;
        }
        inverser_liste_sommets(&(lah->voisins[s]));
        inverser_liste_sommets(&(lal->voisins[s]));
    }
}

liste_triangles calcul_triangle_3(liste_adjacence* la) {
    liste_sommets ll = NULL, lh = NULL;
    liste_adjacence *lal = creer_liste_adjacence(la->nb_sommets);
    liste_adjacence *lah = creer_liste_adjacence(la->nb_sommets);
    liste_triangles lt = creer_liste_triangles();
    lourd_leger(la, &ll, &lh);
    liste_lourd_leger(la, ll, lh, lal, lah);
    for (sommet i = 0; i < la->nb_sommets; ++i)
        for (liste_sommets li = la->voisins[i] ; li ; li = li->suivant)
            if (li->sommet > i)
                intersection(lah, &lt, i, li->sommet);
    for (liste_sommets ls = ll ; ls ; ls = ls->suivant)
        for (liste_sommets li = lal->voisins[ls->sommet] ; li ; li = li->suivant)
            if (li->sommet > ls->sommet)
                intersection(lal, &lt, ls->sommet, li->sommet);
    detruire_liste_sommets(ll);
    detruire_liste_sommets(lh);
    detruire_liste_adjacence(lal);
    detruire_liste_adjacence(lah);
    return lt;
}

void print_triangle(triangle* t) {
    printf("(%d, %d, %d)", t->s0 + 1, t->s1 + 1, t->s2 + 1);
}

void print_liste_triangles(liste_triangles lt) {
    printf("Liste de triangles : ");
    for (; lt; lt = lt->suivant) {
        print_triangle(lt->triangle);
        if(lt->suivant)
            printf(", ");
    }
    printf("\n");
}
